/*
 * Copyright (c) 2008, Stefan Walter
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *     * Redistributions of source code must retain the above
 *       copyright notice, this list of conditions and the
 *       following disclaimer.
 *     * Redistributions in binary form must reproduce the
 *       above copyright notice, this list of conditions and
 *       the following disclaimer in the documentation and/or
 *       other materials provided with the distribution.
 *     * The names of contributors to this software may not be
 *       used to endorse or promote products derived from this
 *       software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include "soa.h"

#include <assert.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

#define SEPARATORS " \t\n\v()"

int
soa_serial_increment (const char *old_soa, char *new_soa, unsigned int new_len)
{
	unsigned long value;
	const char *p, *beg;
	char *end;
	unsigned int len;

	assert (old_soa);
	assert (new_soa);

	/* Check that we can add a digit */
	len = strlen (old_soa);
	assert (new_len >= len + 2);

	p = old_soa;

	/* Eat up space at the beginning */
	while (*p && strchr (SEPARATORS, *p))
		++p;
	/* Skip over domain name, to next sep */
	while (*p && !strchr (SEPARATORS, *p))
		++p;
	/* Eat up space between */
	while (*p && strchr (SEPARATORS, *p))
		++p;
	/* Skip over administrator email, to next sep */
	while (*p && !strchr (SEPARATORS, *p))
		++p;
	/* Eat up spaces between */
	while (*p && strchr (SEPARATORS, *p))
		++p;

	if (*p == 0 || !isdigit (*p)) {
		errno = EINVAL;
		return -1;
	}

	/* Parse the integer here */
	beg = p;
	value = strtoul (beg, &end, 10);
	if (beg == end || !strchr (SEPARATORS, *end)) {
		errno = EINVAL;
		return -1;
	}

	/* Increment the integer appropriately */
	if (value == 0xFFFFFFFF)
		value = 1; /* Zero is recommended against in RFC 1982 */
	else
		++value;

	/* Build up the new SOA */
	memset (new_soa, 0, new_len);
	len = beg - old_soa;
	strncpy (new_soa, old_soa, len);
	snprintf (new_soa + len, 32, "%lu", value);
	len = strlen (new_soa);
	strcpy (new_soa + len, end);

	len = strlen (new_soa);
	assert (len < new_len);

	return len;
}
